//---------------------------------------------------------------------------
//	Conditional Filter for VirtualDub - implements simple branching
//	Copyright (C) 2001-2003 Dmitri Schamschurko
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//      The author can be contacted at:
//      Dmitri Schamschurko
//      dc1@tcs.inf.tu-dresden.de
//
//      Version 1.0     Oktober 2001
//      Version 2.0     March 2003
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "f_ConditionData.h"
#include "f_ConditionLogic.h"

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
VFBitmapInfo::VFBitmapInfo(const VFBitmap& b)
{
  depth   = b.depth;
  w       = b.w;
  h       = b.h;
  pitch   = b.pitch;
  modulo  = b.modulo;
  size    = b.size;
  offset  = b.offset;
  dwFlags = b.dwFlags;
}

void VFBitmapInfo::Restore(VFBitmap& b) const
{
  b.depth   = depth;
  b.w       = w;
  b.h       = h;
  b.pitch   = pitch;
  b.modulo  = modulo;
  b.size    = size;
  b.offset  = offset;
  b.dwFlags = dwFlags;
}

bool VFBitmapInfo::operator == (const VFBitmapInfo& b) const
{
  return
  (b.depth   == depth) &&
  (b.w       == w) &&
  (b.h       == h) &&
  (b.pitch   == pitch) &&
  (b.modulo  == modulo) &&
  (b.size    == size) &&
  (b.offset  == offset) &&
  (b.dwFlags == dwFlags);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
CCondFilterInstance::CCondFilterInstance(FilterActivation* fa, const FilterFunctions *ff)
{
dout(__FUNC__);
  CFD* mfd = *reinterpret_cast<CFD**>(fa->filter_data);
  RefCounter = 1;
  fNoDeinit  = true;
  Started    = false;

  if (!mfd->Filter) return;
dout(__FUNC__ " filter %s found", mfd->Filter->name);

  iSize = mfd->Filter->inst_data_size;
  if (iSize) {
dout(__FUNC__ " allocating %d bytes", iSize);
    pData = new char [iSize];
    if (!pData) {
dout(__FUNC__ " out of memory");
      ff->ExceptOutOfMemory();
    }
    ZeroMemory(pData, iSize);
    fNoDeinit = false;

    if (mfd->Filter->initProc) {
      // fake
      fa->filter      = mfd->Filter;
      fa->filter_data = pData;

dout(__FUNC__ " call initProc (%#x)", fa->filter_data);

      try {
        if (mfd->Filter->initProc(fa, ff)) {
dout(__FUNC__ " initProc failed");
          if (mfd->Filter->deinitProc)
            mfd->Filter->deinitProc(fa, ff);

          delete [] pData;
          pData = 0;
          ff->Except("Conditional: cannot initialize filter instance of '%s'", mfd->Filter->name);
        }
      } __finally {
        // restore
        fa->filter      = mfd->OwnFilter;
        fa->filter_data = mfd->OwnFilterData;
      }
    }
  } else {
    pData = 0;
  }
}
CCondFilterInstance::~CCondFilterInstance()
{
  if (pData) {
dout(__FUNC__ " deallocating memory %#x", pData);
    delete [] pData;
  }
}
void CCondFilterInstance::IncRef()
{
dout(__FUNC__);
  ++RefCounter;
}
void CCondFilterInstance::DeleteInstance(FilterActivation* fa, const FilterFunctions *ff)
{
dout(__FUNC__);
  if (! --RefCounter) {
dout(__FUNC__ " deleting instance...");
    CFD* mfd = *reinterpret_cast<CFD**>(fa->filter_data);
    if (mfd->Filter && mfd->Filter->deinitProc && !fNoDeinit) {
      // check whether there was called endProc
      CallEndProc(fa, ff);

      // fake
      fa->filter      = mfd->Filter;
      fa->filter_data = pData;

dout(__FUNC__ " call deinitProc (%#x)", fa->filter_data);
      mfd->Filter->deinitProc(fa, ff);

      // restore data
      fa->filter      = mfd->OwnFilter;
      fa->filter_data = mfd->OwnFilterData;
    }

dout(__FUNC__ " delete this");
    delete this;
  }
}

void __fastcall CCondFilterInstance::CallStartProc(FilterActivation* fa, const FilterFunctions *ff)
{
  if (!Started) {
    CFD* mfd = *reinterpret_cast<CFD**>(fa->filter_data);

    if (mfd->Filter->startProc)
    try {
      // fake
      fa->filter      = mfd->Filter;
      fa->filter_data = pData;

dout(__FUNC__ " call startProc (%#x)", fa->filter_data);

      mfd->Filter->startProc(fa, ff);
    } __finally {
      // restore data
      fa->filter      = mfd->OwnFilter;
      fa->filter_data = mfd->OwnFilterData;
    }

    Started = true;
  }
}

void __fastcall CCondFilterInstance::CallEndProc(FilterActivation* fa, const FilterFunctions *ff)
{
  if (Started) {
    CFD* mfd = *reinterpret_cast<CFD**>(fa->filter_data);

    if (mfd->Filter->endProc)
    try {
      // fake
      fa->filter      = mfd->Filter;
      fa->filter_data = pData;

dout(__FUNC__ " call endProc (%#x)", fa->filter_data);

      mfd->Filter->endProc(fa, ff);
    } __finally {
      // restore data
      fa->filter      = mfd->OwnFilter;
      fa->filter_data = mfd->OwnFilterData;
    }

    Started = false;
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall CBranch::CallStartProc(FilterActivation* fa, const FilterFunctions *ff)
{
  pInstance->CallStartProc(fa, ff);
  if (Next)
    Next->CallStartProc(fa, ff);
}

void __fastcall CBranch::CallEndProc(FilterActivation* fa, const FilterFunctions *ff)
{
  pInstance->CallEndProc(fa, ff);
  if (Next)
    Next->CallEndProc(fa, ff);
}

CBranch::CBranch(FilterActivation* fa, const FilterFunctions *ff, bool VIsElse)
{
dout(__FUNC__);
  Next = 0;
  pInstance = new CCondFilterInstance(fa, ff);
  IsElse = VIsElse;
  NeedFlags = 0;
}

CBranch::CBranch(const CBranch& Other)
{
dout(__FUNC__ " copy cons");
  if (Other.Next)
    Next = Other.Next->Clone();
  else
    Next = 0;

  pInstance = Other.pInstance;
  if (pInstance)
    pInstance->IncRef();

  IsElse = Other.IsElse;
  NeedFlags = Other.NeedFlags;
}

CBranch::~CBranch()
{
dout(__FUNC__);
  if (Next)
    delete Next;
}

void CBranch::DeleteBranch(FilterActivation* fa, const FilterFunctions *ff)
{
dout(__FUNC__);
  if (pInstance)
    pInstance->DeleteInstance(fa, ff);

  if (Next)
    Next->DeleteBranch(fa, ff);

  Next = 0;
  delete this;
}

bool __fastcall CBranch::HasElse() const
{
  if (IsElse) return true;
  if (Next) return Next->HasElse();
  return false;
}

//---------------------------------------------------------------------------
__fastcall CFormulaObject::~CFormulaObject()
{
  delete pFormula;
  free(pFormulaText);
}
__fastcall CFormulaObject::CFormulaObject(CFormula* Formula, const char* FormulaText)
{
  lRefCounter = 1;
  pFormula = Formula;
  pFormulaText = strdup(FormulaText);
}
void __fastcall CFormulaObject::IncRef()
{
  ++lRefCounter;
}
void __fastcall CFormulaObject::DeleteObject()
{
  if (!--lRefCounter) delete this;
}
bool __fastcall CFormulaObject::Evaluate(CExecutionContext& Context) const
{
  try {
    try {
      return pFormula->Evaluate(Context);
    } catch (...) {}
  } __except(EXCEPTION_EXECUTE_HANDLER) {}
  return false;
}
const char* CFormulaObject::GetFormulaText() const
{
  return pFormulaText;
}
long __fastcall CFormulaObject::GetNeedFlags() const
{
  return pFormula->GetNeedFlags();
}
//---------------------------------------------------------------------------
CIFBranch::CIFBranch(FilterActivation* fa, const FilterFunctions *ff, CFormula* Formula, const char* FormulaText)
  : CBranch(fa, ff, false)
{
  pFormula = new CFormulaObject(Formula, FormulaText);
}

CIFBranch::CIFBranch(const CIFBranch& Other)
  : CBranch(Other)
{
  pFormula = Other.pFormula;
  pFormula->IncRef();
}

CIFBranch::~CIFBranch()
{
  pFormula->DeleteObject();
}

CBranch* CIFBranch::Clone()
{
  return new CIFBranch(*this);
}

CCondFilterInstance* __fastcall CIFBranch::Evaluate(CExecutionContext& Context, int& Which) const
{
  if (pFormula->Evaluate(Context)) return pInstance;
  if (Next) {
    ++Which;
    return Next->Evaluate(Context, Which);
  }
  Which = -1;
  return 0;
}

//---------------------------------------------------------------------------
CELSEBranch::CELSEBranch(FilterActivation* fa, const FilterFunctions *ff)
  : CBranch(fa, ff, true)
{
}
CELSEBranch::CELSEBranch(const CELSEBranch& Other)
  : CBranch(Other)
{
}
CBranch* CELSEBranch::Clone()
{
  return new CELSEBranch(*this);
}
CCondFilterInstance* __fastcall CELSEBranch::Evaluate(CExecutionContext& Context, int& Which) const
{
  return pInstance;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
ConditionalFilterData::ConditionalFilterData()
{
  Filter = 0;
  ff     = 0;

  OwnFilter     = 0;
  OwnFilterData = 0;

  Disabled       = false;
  DisabledReason = 0;
  HaveToSwap     = false;

  Branches      = 0;
  PreviewBranch = 0;

  MonitorPreview = 0;
  MonitorPreviewHandle = 0;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
ExtendedExternalInfo& ExtendedExternalInfo::operator=(ExternalFunctionInfo& Other)
{
  name = Other.name;
  maker = Other.maker;
  description = Other.description;
  fun = Other.fun;

  return *this;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// well, it is not a right place to do it, but...
// for first time i think 100 entries for the external functions should be enought
ExtendedExternalInfo ExternalFunctions[100];

